#!/usr/bin/env perl
use SynBlast;
our $PROGN=[split(/\//, $0)]->[-1];
$VERSION = "SynBlast-" . $VERSION . "\n${PROGN}'s Time-stamp: <2008-08-19 15:07:31 joe>\nsee ${url} for updates.\n";
#######################################################
#doBlastJobs.pl  - part of the SynBlast package for Assisting the Analysis of Conserved Synteny Information
#Copyright (C) 2008  Joerg Lehmann
# 2006-03-29
# joe@bioinf.uni-leipzig.de
#######################################################

=head1 NAME

doBlastJobs.pl - a tool to automatize running tblastn jobs for a selected reference set of query proteins via NCBI BLAST. 
Protein sequences are required locally in fasta format, as well as NCBI BLAST formatted genome databases for the target species of interest.
The default mode is to obtain all blast hits for a species within one program run.

doBlastJobs.pl is part of the L<SynBlast> package for Assisting the Analysis of Conserved Synteny Information.

=head1 SYNOPSIS

B<doBlastJobs.pl> 
[B<-dI<DatabaseName>>]
[B<-sI<AlternativeDatabaseFiles>>]
[B<-eI<QueryDir>>]
[B<-rI<ResultDir>>]
[B<-cI<e-value_CutOff>>]
[B<-tI<addQuotedBlastParam>>]
[B<-n>]
[B<-o>]
[B<-a>]
[B<--help>]
[B<--version>]
protein.info-file

=head1 REQUIRES

require Getopt::Std;
require Pod::Usage;

=head1 DESCRIPTION

There is a second mode available (-a option) where each protein is searched/blasted separately against the database. The resulting files are named, accordingly.
This is done either on free remote hosts accessible with ssh1-key identification (without password), or on a PBS scheduling system [experimental].

=head1 OPTIONS

=over 5

=item B<-d> I<Database_Name> [String]

default = Homo_sapiens

=item B<-e> I<QueryDir> [String]

look for query sequences in QueryDir

default is "infoFile_name" with suffix 'dir'

=item B<-s> I<Alternative_Database_Files> [String]

use alternative db files

default is "/scr/tequilla/genomes/DNA/vertebrata/${DBname}/chr\*.fa"

=item B<-o>


overwrite flag: existing blastresults are ignored
    
=item B<-n>


noPBSflag: ignore available pbs-system, use getFreeHost variant
getFreeHost version needs ssh1-compatible id_keys and set up ssh-agent for ssh without need for password or private key input; to achieve this, the command is: eval `ssh-agent`; ssh-add;

=item B<-a>
each query is blasted separately on a free remote host (using getFreeHost.pl and ssh with key-identification (is always done if -n param set (no pbs)),
or scheduled by an available openPBS queuing system (start from within a pbs server necessary for this)

(default otherwise: all query sequences are searched in one local blast-run)

=item B<-r> I<ResultDir> [String]

write blastresults to ResultDir.
default is {QueryDir}/{BlastProgram_Name}

=item B<-c> I<E-value_CutOff> [String]

default is "1e-10"

=item B<protein.info-file> [String]

The reference info file containing the query names.

=head1 AUTHOR

Joerg Lehmann, <joe@bioinf.uni-leipzig.de>


=head1 AVAILABILITY

http://www.bioinf.uni-leipzig.de/Software/SynBlast/


=head1 LICENSE

This  program is free  software; you can redistribute  it and/or
modify it  under the terms of the GNU  General Public License as
published by the  Free Software Foundation; either version 2, or
(at your option) any later version.

This program is  distributed in the hope that it will be useful,
but WITHOUT  ANY WARRANTY; without even the  implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You can view the  GNU General Public License, online, at the GNU
Project's homepage; see <http://www.gnu.org/licenses/gpl.html>.

See also L<perlgpl> for more information.

=head1  SEE ALSO

L<SynBlast::MySyntenyAln>,
L<SynBlast::MySyntenyUtils>, 
L<SynBlast::MyUtils>,
L<getEnsemblProteins.pl>,
L<doSyntenyFilter.pl>,
L<doSyntenyAlignment.pl>,
L<getBestBlastHit.pl>

=cut

use strict;
use warnings;

use Getopt::Std;
use Pod::Usage;
use SynBlast::MyUtils;
use SynBlast::MySyntenyUtils;
 
$Getopt::Std::STANDARD_HELP_VERSION = 1;
my %opts=();
my $cmdLineWas = $PROGN . " " . join(" ", @ARGV);
getopts('d:p:e:r:s:aoc:nt:h:', \%opts) || pod2usage(-verbose => 0);
 
 pod2usage(-verbose => 0, -message => ("${PROGN}: Error - exactly one InfoFile needed.\n(arguments were: \"@{ARGV}\"\ncmdline was: ${cmdLineWas}\n")) unless @ARGV==1;

my $getHostProg="getFreeHosts.pl -l10";
my $hostprefs="k1 k0 K0";
$hostprefs=$opts{"h"} if(exists($opts{"h"}));
my $pbsFlag=1;
$pbsFlag=0 if(exists($opts{"n"}));

my $tmpPBS = "tmpPBS";

$pbsFlag=0 if($pbsFlag && (system("qstat 2>/dev/null >/dev/null")));

if((exists($opts{"a"})) && ((! `${getHostProg} --help 2>/dev/null`) && (! $pbsFlag))) {
     die "Cannot use either ${getHostProg} nor pbs-queuing system!\nUse option -a instead or start script from pbs-server!\n";
 }

my $blastProg="tblastn";
#$blastProg=$opts{"p"} if(exists($opts{"p"}));

my $DBname="Homo_sapiens";
$DBname=$opts{"d"} if(exists($opts{"d"}));

my $DBdir="/scr/tequilla/genomes/DNA/vertebrata/${DBname}";
my $globStr="${DBdir}/chr*.fa";
$globStr=$opts{"s"} if(exists($opts{"s"}));
print "GlobStr is " . $globStr . "\n";

my @DBfiles= glob($globStr);
if(exists($opts{"s"})) {
    for(my $li=0; $li<@DBfiles; $li++) {
	my @dummyA= split(/\./,$DBfiles[$li]); 
	if($dummyA[-1] eq "log") {
	    pop(@dummyA);
	    $DBfiles[$li] = join(".", @dummyA);
	}
    }
}

my $blEvalue = "1e-10";
$blEvalue = $opts{"c"} if(exists($opts{"c"}));
print "Using e-value of ${blEvalue} for ${blastProg}...\n";

my $addBlParam = "";
$addBlParam = $opts{"t"} if(exists($opts{"t"}));
print "Using additional blast parameter (${addBlParam}) for ${blastProg}...\n";

my $infoFile=$ARGV[0];
my $qSrcDir = ".";
my $DirBase= qx(pwd);
chomp($DirBase);
$qSrcDir=(substr($infoFile, 0, -4) . "dir") if(substr($infoFile, -5) eq ".info");
$qSrcDir=$opts{"e"} if(exists($opts{"e"}));
if(substr($qSrcDir, 0, 1) ne "/") { #srcDir is relative to currentDir?
    $qSrcDir = "${DirBase}/${qSrcDir}";
}
my $bResultDir="${qSrcDir}/${blastProg}";
$bResultDir=$opts{"r"} if(exists($opts{"r"}));
if(substr($bResultDir, 0, 1) ne "/") { #Dir is relative to currentDir?
    $bResultDir = "${DirBase}/${bResultDir}";
}

SynBlast::MyUtils::checkStrEnd(\$qSrcDir, "/") && print "query source directory name changed to " . $qSrcDir . "\n";
SynBlast::MyUtils::checkStrEnd(\$bResultDir, "/") && print "blastresult directory name changed to " . $bResultDir . "\n";

if( -d "${qSrcDir}" ) {
    print "using ${qSrcDir} as source for the sequence files...\n";
} else {
    die "Error: directory \"${qSrcDir}\" (source for the sequence files, -e option) does not exist!\n";
}

if(! (-d "${bResultDir}" )) {
    mkdir( "${bResultDir}" , 0755) or die "Error: Could not create directory \"${bResultDir}\" (-r option, blastresult directory): $!\n";
}
if($pbsFlag && (! (-d "${bResultDir}/${tmpPBS}" ))) {
    mkdir( "${bResultDir}/${tmpPBS}" , 0755) or die "Error: Could not create directory \"${bResultDir}/${tmpPBS}\" (blastresult tmpPBS directory): $!\n";
}

print "using ${bResultDir} as output dir for the blast results...\n";

print "using ${DBname} as database...\n";
print "using ${blastProg} as the blast program option...\n";

# read in infoFile
my @exons=();
my %exonIDIdxHash=();
my @refExons=();
my @fileList=();
SynBlast::MySyntenyUtils::infofile2Exonlist("${infoFile}"
																, \@exons
																, \%exonIDIdxHash
																, \@refExons
																, \@fileList
																, 0
      													, "protID");
print "-> ${infoFile} read. (" . @fileList . " entries)\n";

if(! exists($opts{"a"})) { # process all queries in one local blastrun...
    my $catCmd="cat";
    # prepare cat-command...
    foreach my $file (@fileList) {
	my $querystr = "${qSrcDir}/${file}.fa";
	if(! -f "${querystr}") {
	    print STDERR "\"${querystr}\" was not found!\n";
	} else {
	    $catCmd = $catCmd . " " . $querystr;
	}
    }
    qx(${catCmd} > ${qSrcDir}/${infoFile}.fa);
    if((exists($opts{"o"})) || (! -f "${bResultDir}/${infoFile}.${DBname}.blastresult")) {
	#overwrite flag o set? ignore existing file and run blast...
	if(-f "${bResultDir}/${infoFile}.${DBname}.blastresult") {
	    print STDERR "\"${bResultDir}/${infoFile}.${DBname}.blastresult\" did already exist, will be overwritten...\n";
	}
	print "-> now run blastsearch (${blastProg}) for all " . @fileList . " queries (\"${infoFile}.fa\")...\n";
	## blastall option -F "m S" means Soft masking with SEG (filters only in the word seeding phase)
	my $blastRESULTFILE="${bResultDir}/${infoFile}.${DBname}.blastresult";
	my $blastERRORFILE="${blastRESULTFILE}.ERROR";
	my $blastCMD="blastall -p ${blastProg} -b 500 -F \"m S\" " 
								. "-e " . ${blEvalue} 
								. " -m8 -d \"" . join(" ", @DBfiles) . "\" -i ${qSrcDir}/${infoFile}.fa ${addBlParam}";
								#. " > ${blastRESULTFILE} 2> ${blastERRORFILE}";
	#system("blastall -p ${blastProg} -b 500 -F \"m S\" " 
	#     . "-e " . ${blEvalue} 
	#   . " -m8 -d \"" . join(" ", @DBfiles) . "\" -i ${qSrcDir}/${infoFile}.fa ${addBlParam} > ${bResultDir}/${infoFile}.${DBname}.blastresult 2> ${bResultDir}/${infoFile}.${DBname}.blastresult.ERROR");
	my $cti=localtime(time);
	#system("echo -n -e '### ${blastRESULTFILE}\n### blast search started on ${cti} with command:\n### ${blastCMD}\n' > ${blastRESULTFILE}");
	open(BLRESF, ">", "${blastRESULTFILE}") or die "FATAL ERROR: Could not write result file (${blastRESULTFILE})!\n";
	print BLRESF "### ${blastRESULTFILE}\n### blast search started on ${cti} with command:\n### ${blastCMD} >> ${blastRESULTFILE} 2> ${blastERRORFILE}\n######\n";
	close(BLRESF);
	system("${blastCMD} >> ${blastRESULTFILE} 2> ${blastERRORFILE}");
	print "...finished.\n";
    }
    else {
	print STDERR "\"${bResultDir}/${infoFile}.${DBname}.blastresult\" already exists. Skip this blastsearch...\n";
    }
    
  } else { # -a flag set?
    # process each query separately on free remote host (with nice option)
    ## or on existing pbs server and cluster nodes... (without nice option then)
    my $niceOpt="nice -n19 ";
    $niceOpt=""; ### if($pbsFlag);
    my $scount=0;
    
    my @freeHosts=();
    my $maxRuns=scalar(@fileList);

    foreach my $file (@fileList) {
      $scount++;
      my $querystr = "${qSrcDir}/${file}.fa";
      if(! -f "${querystr}") {
	print STDERR "\"${querystr}\" was not found!\n";
      } else {
	if( exists($opts{"o"}) or (not (( -f "${bResultDir}/${file}.${DBname}.blastcmd.sh" ) or ( -f "${bResultDir}/${file}.${DBname}.blastresult" )))) {
	  #overwrite flag o set? ignore existing file and run blast...
	  if(( -f "${bResultDir}/${file}.${DBname}.blastcmd.sh" ) or ( -f "${bResultDir}/${file}.${DBname}.blastresult")) {
	    print STDERR "\"${bResultDir}/${file}.${DBname}.blastcmd.sh\" and/or its result did already exist, will be overwritten...\n";
	  }

	my $myFreeHost = "";
	if(! $pbsFlag) {
	    
	    my $wtime=5;
	    while(! @freeHosts) {
	      sleep($wtime);
	      foreach my $cHo (split(/\n/, qx(${getHostProg} ${hostprefs}))) {
		push(@freeHosts, $cHo);
	      }
	      if(! @freeHosts) {
		$wtime*= 2;
		print STDERR "waiting for a free host...(${wtime} seconds)\n";
	      }
	    }
	    $myFreeHost=shift(@freeHosts);
	    #print "\n-> host ${myFreeHost} with free capacities found...\n";
	  }
	  
	  open(SCRIPTF, '>', "${bResultDir}/${file}.${DBname}.blastcmd.sh")  or die "Error while creating \"${bResultDir}/${file}.${DBname}.blastcmd.sh\": $!\n";
	  print SCRIPTF "#!/bin/bash\n# blast-script file created by " . $PROGN . "\n";
	  
	  
	  print SCRIPTF 
	      ". /etc/profile;\n",
	      ". ~/.bash_profile;\n", 
	      "cd ${bResultDir} && hostname > ${bResultDir}/${file}.${DBname}.blastHost && ",
	      "${niceOpt}blastall -p ${blastProg} ",

        
	      "-b 100000 ",  ### number of alignments in report is limited to 100000 (default is 250)
	      "-v 100000 ",  ### number of online descriptions to show is limited to 100000 (default is 500)
         
####?????	    "-t 1000 ", ### Length of the largest intron allowed in tblastn for linking HSPs (default/negativeValue is off?)
####???            "-g F ",   ### ungapped style of alignment (default is gapped alignment!)

###### ???      "-f 999 ", ## neighborhood word threshold score (default is 13) -> high value removes neighborhood words and makes seeding require matching words (less sensitiv but extremely faster)
###### ???      "-W 4 ", ## a more stringent word size (default is 3)
	      "-F \"m S\" ", ### Soft masking (with SEG), (soft=>low-complexity filtering only in the word seeding phase, not the extension phase)
	      "-e " . ${blEvalue} . " -m8 ",
	      "-d \"" . join(" ", @DBfiles) . "\" ", #$DBfiles[$li] / $DBfiles[0] 
	      "-i ${querystr} ${addBlParam} > ${bResultDir}/${file}.${DBname}.blastresult 2> ${bResultDir}/${file}.${DBname}.blastresult.ERROR\n";
	  
	  print SCRIPTF 
	      "tmpjoe=`cat ${bResultDir}/${file}.${DBname}.blastresult.ERROR`;\n",
	      "if [ \"\${tmpjoe}\" == \"\" ] ; then\n",
	      "rm ${bResultDir}/${file}.${DBname}.blastresult.ERROR\n",
	      "else\necho `hostname` >> ${bResultDir}/${file}.${DBname}.blastresult.ERROR\nfi\n";
	  close(SCRIPTF);
	  
	  if(! $pbsFlag) {
	    print "-> running ${blastProg} for \"${file}\" (${scount}/${maxRuns}) against ${DBname} on ${myFreeHost}...\n";
	    system("ssh ${myFreeHost} \"bash ${bResultDir}/${file}.${DBname}.blastcmd.sh 2> ${bResultDir}/${file}.${DBname}.blastcmd.ERROR\" &");
	  } else {
	    my @DBname=split(/_/, $DBname);
	    my $shortOrg= substr($DBname[0], 0, 1) . substr($DBname[1], 0, 1);
	    print "-> add ${blastProg} job \"${file}.${shortOrg}\" (${scount}/${maxRuns}) to the pbs queuing system...\n";
	    sleep(2);
	    system("qsub -S /bin/bash -m a -M joe" . '@' . "bioinf.uni-leipzig.de -e ${bResultDir}/${tmpPBS}/${file}.${DBname}.blastcmd.ERROR -o ${bResultDir}/${tmpPBS}/${file}.${DBname}.blastcmd.OUT -N ${file}.${shortOrg} -l walltime=1:0:0 ${bResultDir}/${file}.${DBname}.blastcmd.sh");
	    sleep(2);
	  }
	  
	} else {
	  print "\"${bResultDir}/${file}.${DBname}.blastcmd.sh\" and/or its result  already exists. Skip this blastsearch...\n";
	}
      }
    }
  }

exit 0;

#################################################################################################

sub HELP_MESSAGE {
	pod2usage(-verbose => 1);
}
sub VERSION_MESSAGE {
	print $main::VERSION;
} 

################################################################################
