#!/usr/bin/env perl
use SynBlast;
our $PROGN=[split(/\//, $0)]->[-1];
$VERSION = "SynBlast-" . $VERSION . "\n${PROGN}'s Time-stamp: <2008-08-19 15:00:36 joe>\nsee ${url} for updates.\n";
########################################################
#getEnsemblProteins.pl  - part of the SynBlast package for Assisting the Analysis of Conserved Synteny Information
#Copyright (C) 2008  Joerg Lehmann
# 2006-02-08
# joe@bioinf.uni-leipzig.de
#######################################################

=head1 NAME

getEnsemblProteins.pl - a tool to retrieve syntenic/contiguous sets of protein-coding genes from an Ensembl genome database 
in form of an annotation file (.protein.info suffix) and protein sequence files (fasta format). The gene/protein set is determined 
via a focal gene of interest, together with its flanking protein-coding genes (flanking region as specified by the user).

getEnsemblProteins.pl is part of the L<SynBlast> package for Assisting the Analysis of Conserved Synteny Information.


=head1 SYNOPSIS

B<getEnsemblProteins.pl> 
[B<-fI<FlankingSize>>]
[B<-nI<MaxNumberOfFlankingProteins>>]
[B<-iI<IntervalStart[:IntervalEnd]_position>>]
[B<-lI<contig/chromosome>>]
[B<-dI<ProteinDir>>]
[B<-hI<EnsEMBL_OrganismName>>]
[B<-cI<EnsEMBL_DB_init file>>]
[B<-xI<EnsemblVersionDate>>]
[B<-eI<EnsEMBL_DB_host>>]
[B<-uI<EnsEMBL_DB_user>>]
[B<-s>]
[B<-o>]
[B<--help>]
[B<--version>]
EnsEMBL_ID(s)

=head1 REQUIRES

require Getopt::Std;
require Pod::Usage;
require Bio::EnsEMBL::DBSQL::DBAdaptor;

=head1 DESCRIPTION

getEnsemblProteins.pl retrieves annotation data and protein sequences for a focal gene's flanking region 
from Ensembl Core databases.

=head1 OPTIONS

=over 5

=item B<-f> I<FlankingSize> [String]

flanking size around selected gene, default=1e6

=item B<-i> I<IntervalStart[:IntervalEnd]_position> [String]

additionally, sequence interval can be specified (start and end positions) containing hopefully selected gene, default=not used;
the flanking size param is then used to limit the max size of the interval to search on (started from start position of interval)

=item B<-l> I<chromosome/contig> [String]

given a chromosome/contig name in combination with sequence interval (option -i), it displays only the proteins within the interval and does not save/download anything

=item B<-n> I<MaxNumberOfFlankingProteins> [Integer]

Maximum number of flanking genes for target gene to be used (default=0 => all within flanking size or sequence interval are used).

=item B<-d> I<proteinDir> [String]

save fasta files in proteinDir
default proteinDir is {geneID}_{flankS}.protein.dir

=item B<-h> I<EnsEMBL_OrganismName> [String]

default is Homo_sapiens

the correspondent database for the API release (e.g. 36) is then chosen automatically or as specified in I<EnsEMBL_DB_init file>.

=item B<-c> I<EnsEMBL_DB_init file> [String]

default behaviour => automatic retrieving via 'ensembldb.ensembl.org'

see http://www.ensembl.org/info/software/registry/index.html

=item B<-x> I<Ensembl-DB-Version-String> [String]

e.g. "aug2006" (default: the archive string corresponding to the API version which is in use; normally no change from default necessary)

=item B<-s>


downloads/displays all coding transcripts within region, even if overlapping (for manually editing)
(default is: only the longest protein-transcript for each overlapping transcript groups and strand is downloaded, and further transcripts are only displayed as comment (#))

=item B<-o>


additionally download protein sequences as fasta files (only entries that are not comment out are used)

=item B<EnsEMBL_ID(s)>

The gene ID(s) to use.

=head1 AUTHOR

Joerg Lehmann, <joe@bioinf.uni-leipzig.de>


=head1 AVAILABILITY

http://www.bioinf.uni-leipzig.de/Software/SynBlast/


=head1 LICENSE

This  program is free  software; you can redistribute  it and/or
modify it  under the terms of the GNU  General Public License as
published by the  Free Software Foundation; either version 2, or
(at your option) any later version.

This program is  distributed in the hope that it will be useful,
but WITHOUT  ANY WARRANTY; without even the  implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You can view the  GNU General Public License, online, at the GNU
Project's homepage; see <http://www.gnu.org/licenses/gpl.html>.

See also L<perlgpl> for more information.

=head1  SEE ALSO

L<SynBlast::MySyntenyAln>,
L<SynBlast::MySyntenyUtils>, 
L<SynBlast::MyUtils>,
L<doBlastJobs.pl>
L<doSyntenyFilter.pl>,
L<doSyntenyAlignment.pl>,
L<getBestBlastHit.pl>

=cut

use strict;
use warnings;

use Getopt::Std;
use Pod::Usage;
#use Bio::EnsEMBL::DBSQL::DBAdaptor;
use SynBlast::MyUtils;
use SynBlast::EnsEMBLaccess;
#use DataBrowser;

$Getopt::Std::STANDARD_HELP_VERSION = 1;
my %opts=();
getopts('f:n:d:osc:h:e:u:i:l:x:', \%opts) || pod2usage(-verbose => 0);

my $chromosome = undef;
$chromosome=$opts{"l"} if(exists($opts{"l"}));

# constraints on gene length
#my $maxGL=20000;
#my $minGL=100;
# constraints on length of transcript (maximum number of exons)
#my $maxExons=6;

my $flankS="1e6";
my $flankMP=0;
$flankMP=$opts{"n"} if(exists($opts{"n"}));

my $ensemblVstr=undef; #"dec2006";
$ensemblVstr=lc($opts{"x"}) if(exists($opts{"x"}));
$ensemblVstr = lc(SynBlast::EnsEMBLaccess::getRelMonthString()) unless($ensemblVstr);

my $intvStr=undef;
$intvStr=$opts{"i"} if(exists($opts{"i"}));
my($intvS, $intvE) = (undef, undef);
($intvS, $intvE) = split(/:/, $intvStr) if(defined($intvStr));
if(defined($intvStr)) {
    if(defined($intvE)) { 
	pod2usage(-verbose => 0, -message => "${PROGN}: Error - sequence interval is not correct!\n") unless($intvE>=$intvS);
    }
}

if(! defined($chromosome)) {
    pod2usage(-verbose => 0, -message => "${PROGN}: Error - no EnsEMBL_ID(s) specified.\n") unless @ARGV>0;
} else {
    pod2usage(-verbose => 0, -message => "${PROGN}: Error - no (correct) sequence interval specified for chromosome/contig ${chromosome}!\n") unless(defined($intvE) && ($intvE>=$intvS));
    print "Ignoring geneIDs...\n" if(@ARGV);
}


my $skipFlag=1;
$skipFlag=0 if(exists($opts{"s"}));

my ($host, $user) = (undef, undef); # standard web location (ensembldb.ensembl.org is used automatically, if no initfile set and no other host specified)
$host = $opts{"e"} if(exists($opts{"e"}));
$user = $opts{"u"} if(exists($opts{"u"}));



$flankS = ("" . $opts{"f"}) if(exists($opts{"f"}));
my $saveFlag=0;
$saveFlag = 1 if(exists($opts{"o"}));
my $subDIR;
if(exists($opts{"d"})) {
    $subDIR = $opts{"d"}; 
    SynBlast::MyUtils::checkStrEnd(\$subDIR, "/"); # && print "protein fasta directory name changed to " . $subDIR . "\n";
}

my $initF = undef;
$initF = $opts{"c"} if(exists($opts{"c"}));

my $orgName = "Homo_sapiens";
$orgName = $opts{"h"} if(exists($opts{"h"}));

SynBlast::EnsEMBLaccess::loadRegistry($initF, $host, $user);
my $slice_adaptor = SynBlast::EnsEMBLaccess::getDBadaptor($orgName, "core", "Slice");

if(defined($chromosome)) {
    ### only display interval...
    my $coordsys="";
    $coordsys = "chromosome" if(int($chromosome));
    my $slice = $slice_adaptor->fetch_by_region($coordsys, $chromosome, $intvS, $intvE);
    scanSlice($slice, 0, undef, undef, undef, undef, $intvS, $intvE, $skipFlag);
} else {
    ### display and download interval around specified geneIDs...
    my $gene_adaptor= SynBlast::EnsEMBLaccess::getDBadaptor($orgName, "core", "Gene");
    while(@ARGV) {
	   my $geneID=shift(@ARGV);
	   my $focalgene = $gene_adaptor->fetch_by_stable_id($geneID);
	   if(! defined($focalgene)) {
	    print STDERR "ERROR: Could not fetch ${geneID}, will be skipped!\n";
	    next;
	   }
	   my $fgeneName="";
	   $fgeneName= ($focalgene->external_name() . "_") if(defined($focalgene->external_name()));
	   my $infoF="${fgeneName}${geneID}_${flankS}";
	   $infoF .= "_${intvS}_" . (defined($intvE) ? "${intvE}" : "undef") #. "_${flankMP}" 
	                                     if(defined($intvStr));
     $infoF .= ".protein";
	   $subDIR = "${infoF}.dir" if(! exists($opts{"d"}));
	   my $slice = $slice_adaptor->fetch_by_gene_stable_id($geneID, $flankS);
	   scanSlice($slice, $saveFlag, $geneID, $flankS, $infoF, $subDIR, $intvS, $intvE, $skipFlag);
    }
}
exit 0;
#############################


sub scanSlice {
    my($slice, $saveFlag, $geneID, $flankS, $infoF, $subDIR, $intvS, $intvE, $skipFlag) = @_;
    
    my $scoords = $slice->coord_system()->name();
    my $sseq_region = $slice->seq_region_name();
    my $sstart = $slice->start();
    my $send = $slice->end();
    my $sstrand = $slice->strand();
    my $sline = 
	"#slice: $orgName $scoords $sseq_region $sstart-$send ($sstrand) length=" 
	. $slice->length() #($send - $sstart +1)
	. " bp";
    $sline .= ", ${geneID} + ${flankS} bp flankingSize" if(defined($geneID));
    $sline .= ", interval set to ${intvS}-" if(defined($intvS));
    $sline .= "${intvE}" if(defined($intvE));
    $sline .= "\n";
    print $sline;
 my $headr="#protein.info file\n"
 				 . "#!organism=${orgName}\n"
 				 . "#!ensemblRelease=${ensemblVstr}\n"
 				 . "#!coordsystem=${scoords}\n"
				 . "#!seqregion=${sseq_region}\n"
				 . "#!from=${sstart}\n"
				 . "#!to=${send}\n"
				 . "#!strand=${sstrand}\n"
				 . "#!length=" . $slice->length() . "\n";
		 $headr .= ("#!focalgene=${geneID}\n"
		 					. "#!flankingsize=${flankS}\n") if(defined($geneID));
		 if(defined($intvS)) {
		  		$headr .= "#interval set to ${intvS}-";
		      $headr .= "${intvE}" if(defined($intvE));
		      $headr .= "\n";
		 }
		 
		 
		 my @tcolumns= qw(startPos endPos meanPos protID protName ori geneID geneName transID protLength numExons);
 my $tabhead="#!columns=" . join(",", @tcolumns) . "\n";
 print $headr;
 print $tabhead;
    
    my @sgenes = @{$slice->get_all_Genes()};
    my $countAll = 0;
    my $countHits=0;
    my $countNonCoding=0;
    
    #print all surrounding proteins with location on slice
    if($saveFlag) {
	   # print ProteinInfos in separate File
	   open(INFOFILE, '>', "${infoF}.info")  or die "Error while creating ${infoF}.info: $!\n";
	   #print INFOFILE $sline;
	   print INFOFILE $headr;
	   print INFOFILE $tabhead;
	   (mkdir( $subDIR , 0755) or print STDERR "Error: Could not create directory \"${subDIR}\": $!\n") unless (-d $subDIR);
    }
    
    my @flEntries = ();
    foreach my $sgen (@sgenes) {
      my $gextname = "-";
      $gextname = $sgen->external_name() if(defined($sgen->external_name()));
      my @gtranscripts = @{$sgen->get_all_Transcripts()};
	    foreach my $trans (@gtranscripts) {
	     ++$countAll;
	     my $extname = "-";
	     if (my $extn = $trans->external_name()) {
		    $extname = $extn;
	     }
	    if(! $trans->translation()) {
		     # test on non-coding transcript...
		    ++$countNonCoding;
		    next;
	    }
	    #my $t_length = $trans->length;
	    #	if(! (${t_length} >= ${minGL}) && (${t_length} <= ${maxGL})) {
	    ## sum of all exons (current transcript length) is not in range of specified GeneLength?
	    # next;
	    #}
	    
	    # coding region is not completely  within slice???
	    next if(!(($trans->coding_region_start() > 0) && ($trans->coding_region_end() <= $slice->length()))); 
    	
    	next if(defined($intvS) && ($sstart -1 + $trans->coding_region_start() < $intvS));
	    next if(defined($intvE) && ($sstart -1 + $trans->coding_region_end() > $intvE));
	    
	     
	    my $pep=$trans->translate();
	    ++$countHits;
	    my $curPepLength=$pep->length();
	    my $texonsref = $trans->get_all_translateable_Exons();
	    $texonsref =  [ reverse @{$texonsref} ] if($trans->strand() < 0);
	    
	    ##define relative middle position of cds 
	    my $preDNAl=(int($curPepLength / 2) -1) * 3; 
	    
	    my $numExons= scalar(@{$texonsref});
	    
	    my $foundFl=0;
	    my $meanCdsPos=0;
	    my $oldStart=-1e11;
	    my $sumOfCdsExonLength=0;
	    for(my $li=0; $li<@{$texonsref}; ++$li) {
				die "Error: exons not sorted correctly (get_all_translateable_Exons() method from Bio::EnsEMBL::...Transcript\n" 
				   unless ($oldStart < $texonsref->[$li]->start());
		    $oldStart = $texonsref->[$li]->start();
		#   print "exon from ", $texonsref->[$li]->start(), " to ", $texonsref->[$li]->end(),
		#  " on seq ", $texonsref->[$li]->display_id(), ", strand ", $texonsref->[$li]->strand(), "\n";
		    my $texLength= abs($texonsref->[$li]->end() - $texonsref->[$li]->start()) + 1;
		    $sumOfCdsExonLength+=$texLength;
		    if(! $foundFl) {
		     if( $preDNAl > $texLength) {
		    	$preDNAl-=$texLength;
		     } else {	$foundFl++;
			            $meanCdsPos=$texonsref->[$li]->start() + $preDNAl;
		     }
		    }
	   }
	    #print "sumOfCdsExonLength=" . $sumOfCdsExonLength . "\ncurPepLength*3=" . ($curPepLength*3) . "\n";
	    
	    ###@tcolumns
	    my %curPepLine=( 'geneID' => $sgen->display_id()
	    							 , 'startPos' =>  ($sstart -1 + $trans->coding_region_start()) ## original chromosome coordinates
	    							 , 'endPos' => ($sstart -1 + $trans->coding_region_end())
	    							 , 'meanPos' => ($sstart -1 + ${meanCdsPos})
	    							 , 'protID' => $pep->display_id()
	    							 , 'protName' => ${extname}
	    							 , 'geneName' => ${gextname}
	    							 , 'ori' => $trans->strand()
	    							 , 'transID' => $trans->display_id()
	    							 , 'protLength' => ${curPepLength}
	    							 , 'numExons' => ${numExons}
	    							 );
	  
	    my $curPepSeq=$pep->seq();
	    my $ensemblPepHeader = ">" . $pep->display_id() 
		. " pep:" . lc($trans->status())
		. " ${scoords}:NCBI??:${sseq_region}:"
		. ($sstart -1 + $trans->coding_region_start()) . ":" . ($sstart -1 + $trans->coding_region_end()) . ":" . $trans->strand() 
		. " gene:" . $sgen->display_id() . ":transcript:" . $trans->display_id();
	    
	    ## go through (all) last two flEntries and test on overlapping ??????????????
	    my $ovlFlag=0;
	    for(my $i=SynBlast::MyUtils::max(0,@flEntries - 2); $i< @flEntries; ++$i) {
	    	my @curRegK = keys %{$flEntries[$i]};
		    die "Error: only one key should be there for regionHash (in list flEntries)\n" unless (@curRegK==1);
		    my @curReg= split(/_/, $curRegK[0]);
		    if(SynBlast::MyUtils::getOverlapSize($curPepLine{'startPos'}, $curPepLine{'endPos'}, $curReg[0], $curReg[1] ) > 0) {
		     ### overlapping with previous group of genes/transcripts...
		     ++$ovlFlag;
		     ## get new Range for group...
		      my $newRegEnd=SynBlast::MyUtils::max($curReg[1], $curPepLine{'endPos'});
		      my $newRegStart=SynBlast::MyUtils::min($curReg[0], $curPepLine{'startPos'});
		      my %curHash = %{$flEntries[$i]->{$curRegK[0]}};
		      push(@{$curHash{$curPepLength}}, [ (\%curPepLine, \$curPepSeq, \$ensemblPepHeader) ]);
		     
		      #overwrite old entry with new one...
		      $flEntries[$i] = { ($newRegStart . "_" . $newRegEnd) => \%curHash };
		      ## stop loop here
		      last;
		    }
	    }
	    if(! $ovlFlag) {
		   my %startHash = ();
		   push(@{$startHash{$curPepLength}}, [ (\%curPepLine, \$curPepSeq, \$ensemblPepHeader) ]);
	     push(@flEntries, { ($curPepLine{'startPos'} . "_" . $curPepLine{'endPos'}) => \%startHash });
		  }
	 } ##foreach trans
  } ## foreach sgen
     
    
    ### now write out only non-overlapping protein/trans entries out of flEntries
    #DataBrowser::browse(\@flEntries);
    my $dcounter=0;
    for(my $i=0; $i<@flEntries; ++$i) {
		my @curKe = keys %{$flEntries[$i]};
	
	## sort according to pepLength-key...
	my @cSoKe = reverse sort {$a <=> $b} (keys %{$flEntries[$i]->{$curKe[0]}});
	#print "#--- " . $curKe[0] . " ---#\n";
	##only longest protein-entry for each ovl-group and each strand-possibility is printed normal, others are commented out with #...
	my $posSWritten=0;
	my $negSWritten=0;
	for(my $yy = 0; $yy < @cSoKe; ++$yy) {
	 for(my $zz = 0; $zz < @{$flEntries[$i]->{$curKe[0]}->{$cSoKe[$yy]}}; ++$zz) {
		my @curPepInfo = @{$flEntries[$i]->{$curKe[0]}->{$cSoKe[$yy]}->[$zz]};
		my %curPepLine = %{$curPepInfo[0]};
		
		#my $outLine = join("\t", @curPepLine) . "\n";
		my $outLine =  join("\t", map($curPepLine{$_}, @tcolumns))  . "\n";
		
		++$dcounter;
		if($skipFlag) {
		    if(( ($curPepLine{'ori'} > 0) && $posSWritten) || (($curPepLine{'ori'} < 0) && $negSWritten)) {
			   $outLine= ("# " . $outLine);
			   --$dcounter;
    		}
		 }
		print $outLine;
		
		if($saveFlag > 0) {
		    print INFOFILE $outLine;
		    if( (! $skipFlag) || (! (($curPepLine{'geneID'} ne $geneID) && ((( ($curPepLine{'ori'} > 0) && $posSWritten) || (($curPepLine{'ori'} < 0) && $negSWritten))) ) ) ) {
			##save sequence data for every exon_gene (if not commented out and if skipFlag set)
			my $curPepID=$curPepLine{'protID'};
			my $curPepSeq=${$curPepInfo[1]};
			my $curPepHead=${$curPepInfo[2]};

			if(-f "${subDIR}/${curPepID}.fa") {
			    print "\"${subDIR}/${curPepID}.fa\" did already exist, skip downloading...\n";
			} else {
			    open(SEQFILE, '>', "${subDIR}/${curPepID}.fa")  or die "Error while creating ${subDIR}/${curPepID}.fa: $!\n";
			    print SEQFILE "${curPepHead}\n";
			    ###
  		    print SEQFILE "${curPepSeq}\n";
			    close SEQFILE;
			 }
		  }
		}
		
		++$posSWritten if($curPepLine{'ori'} > 0);
		++$negSWritten if($curPepLine{'ori'} < 0);
	 } ## for zz
	} ## for yy
 }
    
    my $statline = "${dcounter} transcripts passed filter; (in " . ($dcounter - @flEntries) . " cases genes are overlapping, but of different strand" #or within the reference geneID
	. ".)\n# " . (${countHits} - $dcounter) . " transcripts are overlapping with a transcript of same strand but larger/same size, displayed as a comment only (overlappings of referenceID-proteins are all downloaded, but only one longest is displayed).\n# (+ " . (${countAll}-${countHits}) . " skipped entries not displayed, whereof " . $countNonCoding . " nonCoding)\n";
    $statline .= "# on slice around ensembl-ID ${geneID}\n" if(defined($geneID));
    
    if(! $skipFlag) {
	$statline = "${countHits} transcripts (+ " . (${countAll}-${countHits}) . " skipped, whereof " . $countNonCoding . " nonCoding)";
	$statline .= " on slice around ensembl-ID ${geneID}" if(defined($geneID));
	$statline .=  " found/saved.\n";
    }
    
    if($saveFlag > 0) {
	print INFOFILE "# " . $statline;
	close INFOFILE;
    }
    print $statline;
}

#################################################################################################

sub HELP_MESSAGE {
	pod2usage(-verbose => 1);
}
sub VERSION_MESSAGE {
	print $main::VERSION;
} 

################################################################################
